@@ -41,13 +41,17 @@
 #include "panel-object-loader.h"
 #include "panel-schemas.h"
 #include "panel-toplevel.h"
+#include "panel-util.h"
 
 #include "panel-layout.h"
 
 static GSettings *layout_settings = NULL;
 
 #define PANEL_LAYOUT_ERROR panel_layout_error_quark ()
+
+#define PANEL_LAYOUT_DEFAULT_LAYOUT_FILE "panel-default-layout.layout"
 #define PANEL_LAYOUT_OBJECT_GCONF_PATH_TEMPLATE "/apps/panel3-applets/%s"
+#define PANEL_LAYOUT_INSTANCE_CONFIG_SUBPATH "@instance-config/"
 
 static void panel_layout_load_toplevel    (const char *toplevel_id);
 static void panel_layout_load_object      (const char *object_id);
@@ -276,6 +280,81 @@ panel_layout_find_free_id (const char *id_list_key,
         return unique_id;
 }
 
+static gboolean
+panel_layout_maybe_append_object_instance_config (GKeyFile    *keyfile,
+                                                  const char  *group,
+                                                  const char  *key,
+                                                  const char  *path_prefix,
+                                                  const char  *unique_id,
+                                                  gboolean     dry_run,
+                                                  gboolean    *key_handled,
+                                                  GError     **error)
+{
+        char       *value_str;
+        const char *keyname;
+        GVariant   *variant;
+
+        *key_handled = FALSE;
+
+        if (!g_str_has_prefix(key, PANEL_LAYOUT_INSTANCE_CONFIG_SUBPATH))
+                return TRUE;
+
+        *key_handled = TRUE;
+
+        value_str = g_key_file_get_string (
+                                keyfile,
+                                group, key,
+                                error);
+        if (!value_str)
+                return FALSE;
+
+        variant = g_variant_parse (NULL, value_str,
+                                   NULL, NULL, error);
+
+        if (!variant) {
+                g_free (value_str);
+                return FALSE;
+        }
+
+        keyname = key + strlen (PANEL_LAYOUT_INSTANCE_CONFIG_SUBPATH);
+
+        if (dry_run) {
+                /* the key can actually be in a subdirectory
+                 * like instance-config/foo/key, so we split
+                 * the tokens to validate all of them */
+                char **tokens;
+                char **token;
+
+                tokens = g_strsplit (keyname, "/", -1);
+
+                for (token = tokens; *token; token++) {
+                        if (!panel_gsettings_is_valid_keyname (*token,
+                                                               error)) {
+                                g_strfreev (tokens);
+                                g_variant_unref (variant);
+                                g_free (value_str);
+                                return FALSE;
+                        }
+                }
+
+                g_strfreev (tokens);
+        } else {
+                char *key;
+
+                key = g_strdup_printf ("%s%s/%s%s",
+                                       path_prefix, unique_id,
+                                       PANEL_LAYOUT_OBJECT_CONFIG_SUFFIX,
+                                       keyname);
+                panel_dconf_write_sync (key, variant, NULL);
+                g_free (key);
+        }
+
+        g_variant_unref (variant);
+        g_free (value_str);
+
+        return TRUE;
+}
+
 static gboolean
 panel_layout_append_group_helper (GKeyFile                  *keyfile,
                                   const char                *group,
@@ -292,6 +371,7 @@ panel_layout_append_group_helper (GKeyFile                  *keyfile,
                                   const char                *type_for_error_message)
 {
         gboolean    retval = FALSE;
+        gboolean    appending_object;
         const char *id;
         char       *unique_id = NULL;
         char       *path = NULL;
@@ -304,6 +384,8 @@ panel_layout_append_group_helper (GKeyFile                  *keyfile,
 
         g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
+        appending_object = (g_strcmp0 (schema, PANEL_OBJECT_SCHEMA) == 0);
+
         /* Try to extract an id from the group, by stripping the prefix,
          * and create a unique id out of that */
         id = group + strlen (group_prefix);
@@ -339,6 +421,25 @@ panel_layout_append_group_helper (GKeyFile                  *keyfile,
         for (i = 0; keyfile_keys[i] != NULL; i++) {
                 gboolean found = FALSE;
 
+                /* special case keys of the instance config of an object */
+                if (appending_object) {
+                        gboolean handled;
+
+                        if (!panel_layout_maybe_append_object_instance_config (
+                                                        keyfile,
+                                                        group,
+                                                        keyfile_keys[i],
+                                                        path_prefix,
+                                                        unique_id,
+                                                        dry_run,
+                                                        &handled,
+                                                        error))
+                                goto out;
+
+                        if (handled)
+                                continue;
+                }
+
                 for (j = 0; j < key_definitions_len; j++) {
                         if (g_strcmp0 (keyfile_keys[i],
                                        key_definitions[j].name) == 0) {
@@ -473,8 +574,7 @@ panel_layout_append_group (GKeyFile    *keyfile,
 
 static void
 panel_layout_append_from_file_real (const char *layout_file,
-                                    int         screen_for_toplevels,
-                                    gboolean    error_fatal)
+                                    int         screen_for_toplevels)
 {
         GError    *error = NULL;
         GKeyFile  *keyfile = NULL;
@@ -522,9 +622,6 @@ out:
                 g_printerr ("Error while parsing default layout from '%s': %s\n",
                             layout_file, error->message);
                 g_error_free (error);
-
-                if (error_fatal)
-                        g_assert_not_reached ();
         }
 
         if (groups)
@@ -540,14 +637,13 @@ panel_layout_append_from_file_for_screen (const char *layout_file,
 {
         int screen_n = gdk_screen_get_number (screen);
 
-        panel_layout_append_from_file_real (layout_file, screen_n, FALSE);
+        panel_layout_append_from_file_real (layout_file, screen_n);
 }
 
 void
-panel_layout_append_from_file (const char *layout_file,
-                               gboolean    error_fatal)
+panel_layout_append_from_file (const char *layout_file)
 {
-        panel_layout_append_from_file_real (layout_file, -1, error_fatal);
+        panel_layout_append_from_file_real (layout_file, -1);
 }
 
 
@@ -1021,8 +1117,17 @@ panel_layout_load_object (const char *object_id)
 static char *
 panel_layout_get_default_layout_file (void)
 {
+        char *user_file;
+
+        user_file = panel_util_get_from_personal_path (PANEL_LAYOUT_DEFAULT_LAYOUT_FILE);
+
+        if (g_file_test (user_file, G_FILE_TEST_IS_REGULAR))
+                return user_file;
+
+        g_free (user_file);
+
         return g_build_filename (PANELDATADIR,
-                                 "panel-default-layout.layout",
+                                 PANEL_LAYOUT_DEFAULT_LAYOUT_FILE,
                                  NULL);
 }
 
@@ -1095,7 +1200,7 @@ panel_layout_load (void)
                 }
 
                 default_layout_file = panel_layout_get_default_layout_file ();
-                panel_layout_append_from_file (default_layout_file, TRUE);
+                panel_layout_append_from_file (default_layout_file);
                 g_free (default_layout_file);
 
                 toplevels = g_settings_get_strv (layout_settings,
